import java.io.*;
import java.util.ArrayList;
import java.util.List;
import javax.swing.JOptionPane;

public class DeterministicSolver
{
	private String project;
	private int nSpecies;
	private int nReactions;
	private ArrayList<String> speciesNames;
	private int maxReactions = 100;
	private int[][] lefthand;
	private int[][] righthand;
	private int[] reactionType;
	double[] population;
	private double[][] parameters = new double[maxReactions][2];
	double deltaT;
	
	public DeterministicSolver(String din)
	{
		project = din;	
		initialise();
	}
	
	
	private void initialise()
	{
		
		String speciesFile = project + "/species.txt";
		System.out.println(String.format("Loading %s",speciesFile));
		try
		{
			BufferedReader in = new BufferedReader(new FileReader(speciesFile)); 
			speciesNames = new ArrayList<String>();
			String line = "";
			while((line = in.readLine())!=null)
			{
				speciesNames.add(line);	
			}
			nSpecies = speciesNames.size();
			System.out.println(String.format("Read %d species",nSpecies));
			in.close();
		}
		catch(IOException e)
		{
			System.out.println(String.format("Failed to load %s",speciesFile));	
			System.exit(0);
		}
		
		String reactionFile = project + "/reactions.txt";
		System.out.println(String.format("Loading %s",reactionFile));
		
		try
		{
			BufferedReader in = new BufferedReader(new FileReader(reactionFile)); 
			lefthand = new int[maxReactions][nSpecies];
			righthand = new int[maxReactions][nSpecies];
			reactionType = new int[maxReactions];
			String line = "";
			int pos = 0;
			while((line = in.readLine())!=null)
			{
				System.out.println(line);
				String[] components = line.split(":");
				reactionType[pos] = Integer.parseInt(components[0]);
				String[] pars = components[3].split(",");
				if(components[1].length()>0)
				{
					String[] left = components[1].split(",");
					for(int i=0;i<left.length;i++)
					{
						int ind = speciesNames.indexOf(left[i]);
						lefthand[pos][ind] += 1;	
					}
				}
				if(components[2].length()>0)
				{
					String[] right = components[2].split(",");
					for(int i=0;i<right.length;i++)
					{
						int ind = speciesNames.indexOf(right[i]);
						righthand[pos][ind] += 1;	
					}
				}
/*				for(int i=0;i<nSpecies;i++)
				{
					lefthand[pos][i] = Integer.parseInt(left[i]);
					righthand[pos][i] = Integer.parseInt(right[i]);	
				}*/
				for(int i=0;i<pars.length;i++)
				{
					parameters[pos][i] = Double.parseDouble(pars[i]);
				}
				pos ++;
			}
			nReactions = pos;
			in.close();
			System.out.println(String.format("Loaded %d reactions",nReactions));
			for(int i=0;i<nReactions;i++)
			{
				String temp = "";
				for(int j=0;j<nSpecies;j++)
				{
					temp += " " + lefthand[i][j];	
				}
				temp += " : ";
				for(int j=0;j<nSpecies;j++)
				{
					temp += " " +righthand[i][j];	
				}
				System.out.println(temp);				
			}	

		}
		catch(IOException e)
		{
			System.out.println(String.format("Failed to load %s",reactionFile));	
			System.exit(0);
		}
		
		
	
		String initFile = project + "/init.txt";
		System.out.println(String.format("Loading %s",initFile));
		try
		{
			BufferedReader in = new BufferedReader(new FileReader(initFile)); 
			String line = "";
			population = new double[nSpecies];
			for(int i=0;i<nSpecies;i++)
			{
				line = in.readLine();
				population[i] = Double.parseDouble(line);
			}
			System.out.println(String.format("Loaded %s:",initFile));
			System.out.println("Species:");
			for(int i=0;i<nSpecies;i++)
			{
				System.out.println(String.format("%s:%f",speciesNames.get(i),population[i]));	
			}
			in.close();
		}
		catch(IOException e)
		{
			System.out.println(String.format("Failed to load %s",initFile));	
			System.exit(0);	
		}

	}
	private double compGrad(int sNo)
	{
		double g = 0.0;
		for(int i=0;i<nReactions;i++)
		{
			if(reactionType[i]==0) //Mass Action
			{
				if(righthand[i][sNo]>lefthand[i][sNo])
				{
					double temp = 1.0;
					for(int j = 0;j<nSpecies;j++)
					{
						if(lefthand[i][j]==1)
						{
							temp *= population[j];	
						}	
					}
					g += temp*parameters[i][0];
				}
				else if(righthand[i][sNo]<lefthand[i][sNo])
				{
					double temp = 1.0;
					for(int j = 0;j<nSpecies;j++)
					{
						if(lefthand[i][j]==1)
						{
							temp *= population[j];	
						}	
					}
					g -= temp*parameters[i][0];
				}
			}
			else if(reactionType[i]==1) // Michaelis-menten
			{
				if(righthand[i][sNo]>lefthand[i][sNo])
				{
					for(int j=0;j<nSpecies;j++)
					{
						if(lefthand[i][j]==1)
						{
							g += (population[j]*parameters[i][0])/(population[j] + parameters[i][1]);
						}	
					}	
				}
				else if(righthand[i][sNo]<lefthand[i][sNo])
				{
					for(int j=0;j<nSpecies;j++)
					{
						if(lefthand[i][j]==1)
						{
							g -= (population[j]*parameters[i][0])/(population[j] + parameters[i][1]);	
						}	
					}
				}	
			}
			else if(reactionType[i]==2) // Repression
			{
				if(righthand[i][sNo]>lefthand[i][sNo])
				{
					for(int j=0;j<nSpecies;j++)
					{
						if(lefthand[i][j]==1)
						{
							g += (parameters[i][0])/(population[j] + parameters[i][1]);
						}	
					}	
				}
				else if(righthand[i][sNo]<lefthand[i][sNo])
				{
					for(int j=0;j<nSpecies;j++)
					{
						if(lefthand[i][j]==1)
						{
							g -= (parameters[i][0])/(population[j] + parameters[i][1]);	
						}	
					}
				}		
			}
			
		}	
		return g;
	}
	public void solve(double maxT,double deltaT,double recordInterval,String outName)
	{
		PrintWriter out;
		double t = 0.0;
		double nextRecord = 0;
		try
		{
			out = new PrintWriter(new FileWriter(outName)); 
			String head = "T";
			for(int i=0;i<nSpecies;i++)
			{
				head += String.format("\t%s",speciesNames.get(i));	
			}
			out.println(head);
			String outline = String.format("%f",t);
			for(int i=0;i<nSpecies;i++)
			{
				outline += String.format("\t%f",population[i]);	
			}
			out.println(outline);
			nextRecord = t + recordInterval;
			while(t<maxT)
			{
				//Compute gradients
				double[] newPop = new double[nSpecies];
				for(int i=0;i<nSpecies;i++)
				{
					double dS = compGrad(i);	
					newPop[i] = population[i]+ dS*deltaT;
				}	
				// Update the species
				for(int i=0;i<nSpecies;i++)
				{
					population[i] = newPop[i];	
				}
				t+=deltaT;
				if(t>=nextRecord)
				{
					outline = String.format("%f",t);
					for(int i=0;i<nSpecies;i++)
					{
						outline += String.format("\t%f",population[i]);	
					}
					out.println(outline);
					nextRecord = t + recordInterval;
				}
				
			}
			out.close();
		}
		catch(IOException e)
		{
			System.out.println("Failed to open outfile");
			System.exit(0);	
		}
		
	}
		
	public static void main(String[] args)
	{
		String projectDir = "";
		double tempmaxT = 0.0;
		double tempdeltaT = 0.0;
		double tempreportT = 0.0;
		String outName = "";
		if(args.length==5)
		{
			try
			{
				projectDir = args[0];
				tempmaxT = Double.parseDouble(args[1]);
				tempdeltaT = Double.parseDouble(args[2]);
				tempreportT = Double.parseDouble(args[3]);	
				outName = args[4];
			}
			catch(NumberFormatException e)
			{
				System.out.println("Arguments: project maxT deltaT reportT outfile");
				System.exit(0);
			}	
		}
		else if(args.length==0)
		{
			projectDir = (String)JOptionPane.showInputDialog(null,"Enter Project Directory");
			if(projectDir==null)
			{
				System.exit(0);	
			}
			
			while(true)
			{
				String temp = (String)JOptionPane.showInputDialog(null,"Enter maximum simulation time");
				if(temp==null)
				{
					System.exit(0);	
				}
				else
				{
					try
					{
						tempmaxT = Double.parseDouble(temp);	
					}
					catch(NumberFormatException e)
					{
						continue;
					}
					break;
				}
			}
			
			while(true)
			{
				String temp = (String)JOptionPane.showInputDialog(null,"Enter deltat for solver");
				if(temp==null)
				{
					System.exit(0);	
				}
				else
				{
					try
					{
						tempdeltaT = Double.parseDouble(temp);	
					}
					catch(NumberFormatException e)
					{
						continue;
					}
					break;
				}
			}
			
			while(true)
			{
				String temp = (String)JOptionPane.showInputDialog(null,"Enter reportt");
				if(temp==null)
				{
					System.exit(0);	
				}
				else
				{
					try
					{
						tempreportT = Double.parseDouble(temp);	
					}
					catch(NumberFormatException e)
					{
						continue;
					}
					break;
				}
			}
			
			outName = (String)JOptionPane.showInputDialog(null,"Enter output file name");
			if(outName == null)
			{
				System.exit(0);	
			}
			
		}
		else
		{
			System.out.println("Arguments: project maxT deltaT reportT outfile");
			System.exit(0);		
		}
		DeterministicSolver ma = new DeterministicSolver(projectDir);
		ma.solve(tempmaxT,tempdeltaT,tempreportT,outName);
		
	}
}